#!/usr/bin/env python

"""
Module for representing
  - complex nested list or tuple or dictionary
  - as a Tree based data structure (termed as Scatter Tree)

Tree reprentation allows
  - equality checks on contents
  - even though the contents have different order
"""

from enum import Enum

class NodeType(Enum):
    """
    Enumeration Class for Type of Node in a Scatter Tree
    """
    PRIMITIVE = 1
    LIST_OR_TUPLE = 2
    DICTIONARY = 3
    KEY_VALUE = 4

class Marker(Enum):
    """
    Enumeration Class to Flag a Scatter Tree Node
      - GREEN, when a node is not matched yet
      - RED, when a node has been matched
    """
    RED = False
    GREEN = True

class Node:
    """
    Basic Node class.
    Actual Nodes of a Scatter tree will inherit this class.

    Structure:
        +------------+
        |    type    |
        |------------|
        |   marker   |
        +------------+
    """
    def __init__(self, node_type):
        if node_type in NodeType:
            self.type = node_type
        else:
            raise TypeError("Invalid value for 'node_type'")
        self.marker = Marker.GREEN
    def get_type(self):
        """
        Getter method for attribute 'type'
        """
        return self.type
    def get_marker(self):
        """
        Getter method for attribute 'marker'
        """
        return self.marker
    def set_marker(self):
        """
        Setter method for attribute 'marker'
        """
        if self.marker is Marker.GREEN:
            self.marker = Marker.RED
        else:
            raise ValueError("Flag is already marked. Cannot mark again.")

class PrimitiveNode(Node):
    """
    Class to denote a Node representing Primitive Datatype.

    Structure:
        +------------+
        |    type    |
        |------------|
        |   marker   |
        |------------|
        |   value    |
        +------------+
    """
    def __init__(self, value):
        super().__init__(NodeType.PRIMITIVE)
        self.value = value
    def get_value(self):
        """
        Getter method for attribute 'value'
        """
        return self.value

class CollectionNode(Node):
    """
    Class to denote a Node representing a Collection Datatype.

    Structure:
        +------------+
        |    type    |
        |------------|
        |   marker   |
        |------------|
        |    size    |
        |------------|
        |  members   |
        +------------+
    """
    def __init__(self, collectionType, size):
        if collectionType in (NodeType.LIST_OR_TUPLE, NodeType.DICTIONARY):
            super().__init__(collectionType)
        else:
            raise TypeError("Invalid value for 'collectionType'")
        self.size = size
        self.members = []
    def get_size(self):
        """
        Getter method for attribute 'size'
        """
        return self.size
    def get_members(self):
        """
        Getter method for attribute 'members'
        """
        if len(self.members) != self.size:
            raise ValueError("Attribute 'members' is incomplete")
        return self.members
    def add_member(self, node):
        """
        Setter method for attribute 'members'
        """
        if len(self.members) == self.size:
            raise OverflowError("cannot add members more than the initialized size")
        self.members.append(node)
        return True

class ListNode(CollectionNode):
    """
    Class to denote a Node representing a List or Tuple.

    Structure:
        +------------+
        |    type    |
        |------------|
        |   marker   |
        |------------|
        |    size    |
        |------------|
        |  members   |
        +------------+
    """
    def __init__(self, size):
        super().__init__(NodeType.LIST_OR_TUPLE, size)
    def filter_collection_members(self, member_type, member_length):
        """
        Method to filter selective 'members'
          which are collections with known length
        """
        if member_type not in (NodeType.LIST_OR_TUPLE, NodeType.DICTIONARY):
            raise TypeError("Invalid value for 'member_type'")
        match = [ m for m in self.get_members()
                      if m.get_type() is member_type
                          and m.get_size() == member_length
                          and m.get_marker()
                ]
        return match
    def filter_primitive_members(self, member_value):
        """
        Method to filter selective 'members'
          which are primitive with known value
        """
        match = [ m for m in self.get_members()
                      if m.get_type() is NodeType.PRIMITIVE
                          and m.get_value() == member_value
                          and m.get_marker()
                ]
        return match

class DictNode(CollectionNode):
    """
    Class to denote a Node representing a Dictionary.

    Structure:
        +------------+
        |    type    |
        |------------|
        |   marker   |
        |------------|
        |    size    |
        |------------|
        |  members   |
        +------------+
    """
    def __init__(self, size):
        super().__init__(NodeType.DICTIONARY, size)
    def filter_similar_members(self, sample_member):
        """
        Method to filter selective 'members'
          which are similar to a given key value node
        """
        #
        key_type = sample_member.get_key_reference().get_type()
        if key_type is NodeType.PRIMITIVE:
            is_key_collection = False
            key_filter = sample_member.get_key_reference().get_value()
        else:
            is_key_collection = True
            key_filter = sample_member.get_key_reference().get_size()
        #
        value_type = sample_member.get_value_reference().get_type()
        if value_type is NodeType.PRIMITIVE:
            is_value_collection = False
            value_filter = sample_member.get_value_reference().get_value()
        else:
            is_value_collection = True
            value_filter = sample_member.get_value_reference().get_size()
        #
        # filter matching for K node type and V node type
        both_types_match = [ m for m in self.get_members()
                                 if m.get_key_reference().get_type() == key_type
                                   and m.get_value_reference().get_type() == value_type
                           ]
        # filter matching for K node size or K node value
        only_key_match = [ m for m in both_types_match
                               if (  is_key_collection
                                     and m.get_key_reference().get_size() == key_filter
                                  )
                                  or (  (not is_key_collection)
                                        and m.get_key_reference().get_value() == key_filter
                                     )
                         ]
        # filter matching for V node size or V node value
        match = [ m for m in only_key_match
                      if (  is_value_collection
                            and m.get_value_reference().get_size() == value_filter
                         )
                         or (  (not is_value_collection)
                               and m.get_value_reference().get_value() == value_filter
                            )
                ]
        return match

class KeyValNode(Node):
    """
    Class to denote a Node representing a Key Value pair.

    Structure:
        +-------------------+
        |        type       |
        |-------------------|
        |       marker      |
        |-------------------|
        |   key_reference   |
        |-------------------|
        |  value_reference  |
        +-------------------+
    """
    def __init__(self):
        super().__init__(NodeType.KEY_VALUE)
        self.key_reference = None
        self.value_reference = None
    def get_key_reference(self):
        """
        Getter method for attribute 'key_reference'
        """
        return self.key_reference
    def get_value_reference(self):
        """
        Getter method for attribute 'value_reference'
        """
        return self.value_reference
    def set_key_reference(self, node):
        """
        Setter method for attribute 'key_reference'
        """
        self.key_reference = node
        return True
    def set_value_reference(self, node):
        """
        Setter method for attribute 'value_reference'
        """
        self.value_reference = node
        return True



#!/usr/bin/env python

"""
Module for
    1. Translating complex nested datatype
       into a Tree like structure (termed as Scatter Tree)
    2. Comparing two scatter trees for equality

Tree based comparison allows
  - equality checks on contents
  - even though the contents have different order
"""

from .scatter_tree_nodes import NodeType
from .scatter_tree_nodes import ListNode
from .scatter_tree_nodes import DictNode
from .scatter_tree_nodes import KeyValNode
from .scatter_tree_nodes import PrimitiveNode

def _scatter(iterable_object, node_count=0):
    """
    For internal use only.
    Users are discouraged to invoke this function directly.

    Function to create a scatter tree from an object.
    It recursively parses the given 'iterable_object'
      - creates a node for each element
      - builds the scatter tree
      - it also tracks the total number of nodes.

    Parameters:
        Name:  iterable_object
        Type:    Any
        Name:  node_count    (OPTIONAL)
        Type:    int         (default is 0)

    Return Type:
       tuple(object_reference, int)
           'object_reference' points to Root Node of the Scatter Tree
           'int' value denotes the total node count in the Scatter Tree
    """
    if isinstance(iterable_object, (list, tuple)):
        # scatter list/tuple into many nodes
        list_node = ListNode(len(iterable_object))
        node_count += 1
        for m_obj in iterable_object:
            m_node, node_count = _scatter(m_obj, node_count)
            list_node.add_member(m_node)
        root_node = list_node
    elif isinstance(iterable_object, dict):
        # scatter dict into many nodes
        dict_node = DictNode(len(iterable_object))
        node_count += 1
        for (k_obj,v_obj) in iterable_object.items():
            k_node, node_count = _scatter(k_obj, node_count)
            v_node, node_count = _scatter(v_obj, node_count)
            kv_node = KeyValNode()
            node_count += 1
            kv_node.set_key_reference(k_node)
            kv_node.set_value_reference(v_node)
            dict_node.add_member(kv_node)
        root_node = dict_node
    else:
        # retain as single node
        primitive_node = PrimitiveNode(iterable_object)
        node_count += 1
        root_node = primitive_node
    return (root_node, node_count)

def scatter(iterable_object):
    """
    Function to create a scatter tree from an iterable object.

    Parameters:
        Name:  iterable_object
        Tyoe:    Any

    Return Type:
       tuple(object_reference, int)
           'object_reference' points to Root Node of the Scatter Tree
           'int' value denotes the total node count in the Scatter Tree
    """
    scatter_tree, total_node_count = _scatter(iterable_object)
    return (scatter_tree, total_node_count)

def _compare_primitive_nodes(left, right, depth):
    if left.get_value() != right.get_value():
        # unequal values
        return False
    #
    # mark flags of 'left' and 'right'
    #
    if depth == 1:
        left.set_marker()
        right.set_marker()
    #
    # 'left' and 'right' have matched successfully
    #
    return True

def _compare_keyval_nodes(left, right, depth):
    are_keys_similar = _compare_nodes_func[left.get_key_reference().get_type()](
                           left.get_key_reference(),
                           right.get_key_reference(),
                           depth+1
                       )
    if are_keys_similar is False:
        # keys are different
        return False
    are_values_similar = _compare_nodes_func[left.get_value_reference().get_type()](
                             left.get_value_reference(),
                             right.get_value_reference(),
                             depth+1
                         )
    if are_values_similar is False:
        # values are different
        return False
    #
    # mark flags of 'left' and 'right'
    #
    if depth == 1:
        _mark_recursively(left)
        _mark_recursively(right)
    #
    # 'left' and 'right' have matched successfully
    #
    return True

def _compare_collection_nodes(left, right, depth):
    cache_nodes = list()
    # iterate over 'left'
    for member in left.get_members():
        #
        # member can be list node / dict node / primitive node
        #
        # search for similar match in 'right'
        probable_matches = _fetch_similar_nodes(right, member)
        if len(probable_matches) == 0:
            # didn't find a match
            return False
        # iterate over many matches
        is_matched = False
        for match in probable_matches:
            # skip if it was matched before
            if match in cache_nodes:
                continue
            # otherwise compare recursively
            if _compare_nodes_func[member.get_type()](member, match, depth+1):
                # 'left'::'member' and 'right'::'match' are equal
                # stop iterating matches
                cache_nodes.append(match)
                is_matched = True
                break
        #<------
        if is_matched is False:
            # didn't find any match
            return False
        #
        # mark flags of 'member' and last item in 'cache_nodes'
        #
        if depth == 1:
            _mark_recursively(member)
            _mark_recursively(cache_nodes[-1])
    #
    # 'left' and 'right' have matched successfully
    #
    return True

def _fetch_similar_nodes(scatter_tree, sample_node):
    # In List
    if scatter_tree.get_type() is NodeType.LIST_OR_TUPLE:
        if sample_node.get_type() in (NodeType.LIST_OR_TUPLE, NodeType.DICTIONARY):
            # filter collections
            similar_nodes = scatter_tree.filter_collection_members( sample_node.get_type(),
                                                                    sample_node.get_size()
                                                                  )
        elif sample_node.get_type() is NodeType.PRIMITIVE:
            # filter primitives
            similar_nodes = scatter_tree.filter_primitive_members(sample_node.get_value())
    # In Dict
    if scatter_tree.get_type() is NodeType.DICTIONARY:
        if sample_node.get_type() is NodeType.KEY_VALUE:
            # filter similar KV nodes
            similar_nodes = scatter_tree.filter_similar_members(sample_node)
    return similar_nodes

def _mark_recursively(scatter_tree):
    node_type = scatter_tree.get_type()
    # set marker on child nodes
    if node_type in (NodeType.LIST_OR_TUPLE, NodeType.DICTIONARY):
        for member in scatter_tree.get_members():
            _mark_recursively(member)
    elif node_type is NodeType.KEY_VALUE:
        _mark_recursively(scatter_tree.get_key_reference())
        _mark_recursively(scatter_tree.get_value_reference())
    # set marker on input node
    scatter_tree.set_marker()

# mapping node types against their respective compare function
_compare_nodes_func = {NodeType.PRIMITIVE : _compare_primitive_nodes,
                       NodeType.LIST_OR_TUPLE : _compare_collection_nodes,
                       NodeType.DICTIONARY : _compare_collection_nodes,
                       NodeType.KEY_VALUE : _compare_keyval_nodes
                      }

def check_scatter_equality(scatter_left, scatter_right):
    """
    Function to check if 2 scatter trees are equivalent.

    Parameters:
        Name: scatter_left
        Type:    scatter_tree_nodes.ListNode
                 scatter_tree_nodes.DictNode
                 scatter_tree_nodes.KeyValNode
                 scatter_tree_nodes.PrimitiveNode
        Name: scatter_right
        Type:    scatter_tree_nodes.ListNode
                 scatter_tree_nodes.DictNode
                 scatter_tree_nodes.KeyValNode
                 scatter_tree_nodes.PrimitiveNode

    Return Type:
        Boolean
    """
    if ( ( not isinstance(scatter_left, scatter_right.__class__) )
         or
         ( not isinstance( scatter_left,
                             (ListNode, DictNode, PrimitiveNode, KeyValNode)
                         )
         )
       ):
        # different input types
        # or unexpected input type
        return False
    # compare and return the outcome
    return _compare_nodes_func[scatter_left.get_type()](
                                                  scatter_left,
                                                  scatter_right,
                                                  0
                                                )

def compare_two_collections(first_collection, second_collection):
    """
    Function to compare 2 complex nested collections.
    Input Collection can be a tuple / list / dict.
    It translates the collection into a scatter tree
      and then compare the corresponding scatter trees

    Parameters:
        Name: first_collection
        Type:    tuple / list / dict
        Name: second_collection
        Type:    tuple / list / dict

    Return Type:
        Boolean
    """
    try:
        first_scatter_tree, first_node_count = scatter(first_collection)
        second_scatter_tree, second_node_count = scatter(second_collection)
        if first_node_count != second_node_count:
            # different node counts
            return False
        are_equal = check_scatter_equality(first_scatter_tree, second_scatter_tree)
        return are_equal
    except (TypeError, ValueError, OverflowError):
        return False



#!/usr/bin/env python3

"""
Module to Generate Variance Report
  - uses Pandas library
  - generates Excel file

"""

from pyspark.sql.functions import sum     # pylint: disable=import-error, redefined-builtin
from pyspark.sql.functions import desc    # pylint: disable=import-error
from pandas import ExcelWriter            # pylint: disable=import-error
from .effulge import summarize_variance    # pylint: disable=no-name-in-module


def save_variance_report(variance_df, source_df, target_df, super_key, file_name, **kwargs):    # pylint: disable=too-many-locals
    """
    Function to create Excel report from Effulge Variance Dataframe

    Parameters:
        Name: variance_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: source_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: target_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: super_key
        Type:    tuple or list of String
        Name: file_name
        Type:    String
        Name: src_prefix    (OPTIONAL)
        Type:    String     (default is 'src')
        Name: tgt_prefix    (OPTIONAL)
        Type:    String     (default is 'tgt')

    Return Type:
        None
    """
    # assess impact of each attribute
    _r_summary = summarize_variance(variance_df)
    _r_impacted_attributes = _sort_by_most_impacted_attributes( _r_summary )
    # list out all attributes causing variance
    _list_of_rows = _r_impacted_attributes\
        .where("_attribute_name not in ('MISSING_PRIMARY_KEY','DUPLICATE_PRIMARY_KEY')")\
        .select("_attribute_name").collect()
    _list_of_attributes = [ x.asDict()['_attribute_name'] for x in _list_of_rows ]
    # set Excel configurations to avoid data conversion
    excel_options = {}
    excel_options['strings_to_formulas'] = False
    excel_options['strings_to_urls'] = False
    excel_engine_kwargs = {'options':excel_options}
    excel_writer = ExcelWriter(
        file_name + '.xlsx',
        engine='xlsxwriter',
        engine_kwargs=excel_engine_kwargs
    )
    # create Excel sheet for 'MISSING_PRIMARY_KEY'
    _missing_spark_df = variance_df\
        .where("ARRAY_CONTAINS( EFFULGE_VARIANCE_PROVOKER , 'MISSING_PRIMARY_KEY' )")\
        .select( *super_key )
    _missing_panda_df = _missing_spark_df.toPandas()
    _missing_panda_df.to_excel(excel_writer, sheet_name='MISSING_PRIMARY_KEY', index=False)
    #
    # create Excel sheet for 'DUPLICATE_PRIMARY_KEY'
    _duplicate_spark_df = variance_df\
        .where("ARRAY_CONTAINS( EFFULGE_VARIANCE_PROVOKER , 'DUPLICATE_PRIMARY_KEY' )")\
        .select( *super_key )
    _duplicate_panda_df = _duplicate_spark_df.toPandas()
    _duplicate_panda_df.to_excel(excel_writer, sheet_name='DUPLICATE_PRIMARY_KEY' , index=False)
    #
    # create Excel sheet for varying attributes
    for column in _list_of_attributes :
        _focused_spark_df = _focused_variance_for_each_attribute(variance_df, source_df, target_df,
                                                                 super_key, column, **kwargs)
        _focused_panda_df = _focused_spark_df.toPandas()
        _focused_panda_df.to_excel(excel_writer, sheet_name=column, index=False)
    # save Excel file
    excel_writer.save()


def _sort_by_most_impacted_attributes(variance_df):
    """
    Function to identify impact of each column
    and sort them based on their severity

    Parameters:
        Name: variance_df
        Type:    pyspark.sql.dataframe.DataFrame

    Return Type:
        pyspark.sql.dataframe.DataFrame
        Format:

    +-------------------+----------------+
    |  _attribute_name  |  _total_count  |
    +-------------------+----------------+
    |                   |                |
    |                   |                |
    +-------------------+----------------+

    """
    out_df = variance_df.selectExpr(
            "explode(EFFULGE_VARIANCE_PROVOKER) as _attribute_name",
            "EFFULGE_VARIANCE_COUNT as _count"
        )\
        .groupBy("_attribute_name")\
        .agg(sum("_count").alias("_total_count"))\
        .orderBy( desc("_total_count") )
    #
    return out_df


def _focused_variance_for_each_attribute(
            #pylint: disable=too-many-arguments
            variance_df, source_df, target_df, primary_key_list,
            attribute_name, src_prefix='src', tgt_prefix='tgt'):
    #pylint: disable=line-too-long
    """
    Function to fetch all variances for a given column.

    Parameters:
        Name: variance_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: source_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: target_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: primary_key_list
        Type:    tuple or list of String
        Name: attribute_name
        Type:    String
        Name: src_prefix    (OPTIONAL)
        Type:    String     (default is 'src')
        Name: tgt_prefix    (OPTIONAL)
        Type:    String     (default is 'tgt')

    Return Type:
        pyspark.sql.dataframe.DataFrame
        Format:

    +------------------------+---------+------------------------+--------------------------------------+--------------------------------------+
    |  <<primary_column_1>>  |  . . .  |  <<primary_column_N>>  |  _<<src_prefix>>_<<attribute_name>>  |  _<<tgt_prefix>>_<<attribute_name>>  |
    +------------------------+---------+------------------------+--------------------------------------+--------------------------------------+
    |                        |         |                        |                                      |                                      |
    |                        |         |                        |                                      |                                      |
    +------------------------+---------+------------------------+--------------------------------------+--------------------------------------+

    """
    _filtered_records = variance_df\
        .where(f"ARRAY_CONTAINS( EFFULGE_VARIANCE_PROVOKER, '{attribute_name}')")
    focused_df = _filtered_records\
        .join(
            source_df.selectExpr(*primary_key_list,
                                 f"{attribute_name} as _{src_prefix}_{attribute_name}")  ,
            [ *primary_key_list ] ,
            "left"
        ).join(
            target_df.selectExpr(*primary_key_list,
                                 f"{attribute_name} as _{tgt_prefix}_{attribute_name}")  ,
            [ *primary_key_list ] ,
            "left"
        ).selectExpr(
            *primary_key_list,
            f"_{src_prefix}_{attribute_name}",
            f"_{tgt_prefix}_{attribute_name}"
        )
    #
    return focused_df



#!/usr/bin/env python3

"""
Data Validation Utility on PySpark.

Effulge means "to shine".

Analogy:
  To visually compare any two identical real world objects,
      we need a light source to equip our eyes.
  Similarly, to compare identical pyspark dataframes,
      Effulge utility can equip us to validate with ease.
"""


from pyspark.sql import SparkSession                        # pylint: disable=import-error
from pyspark.sql import Row                                 # pylint: disable=import-error
from pyspark.sql import DataFrame                           # pylint: disable=import-error
from pyspark.sql.types import ArrayType                     # pylint: disable=import-error
from pyspark.sql.types import StringType                    # pylint: disable=import-error
from pyspark.sql.functions import col                       # pylint: disable=import-error
from pyspark.sql.functions import round as pyspark_round    # pylint: disable=import-error
from .scatter_tree_compare import compare_two_collections


def spot_variance(expected_dataframe, available_dataframe, primary_key):
    """
    Function to find column level variance between two dataframes.
    # Assumption:
    #  1. All columns in first dataframe should be present in second dataframe
    #  2. Such common columns must have same column names on both dataframes
    #  3. Second dataframe could contain extra columns or records, they will be ignored

    Parameters:
        Name: expected_dataframe
        Type:    pyspark.sql.dataframe.DataFrame
        Name: available_dataframe
        Type:    pyspark.sql.dataframe.DataFrame
        Name: primary_key
        Type:    tuple or list

    Return Type:
        pyspark.sql.dataframe.DataFrame
        Format:

    +------------------------+---------+------------------------+-----------------------------+
    |  <<primary_column_1>>  |  . . .  |  <<primary_column_N>>  |  EFFULGE_VARIANCE_PROVOKER  |
    +------------------------+---------+------------------------+-----------------------------+
    |                        |         |                        |                             |
    |                        |         |                        |                             |
    +------------------------+---------+------------------------+-----------------------------+

        EFFULGE_VARIANCE_PROVOKER
            Type:    List of Strings
            Values:  Column Names, Constant Message
                Constant Message can be -
                    "MISSING_PRIMARY_KEY"
                        => when 'expected_dataframe' contains a primary key value
                        => but 'available_dataframe' does not have corresponding match
                    "DUPLICATE_PRIMARY_KEY"
                        => when 'expected_dataframe' contains a unique primary key value
                        => but 'available_dataframe' has many matches for the same primary key value

    Throws Exception:
        - if any input parameter value is not of the acceptable type
        - if 'primary_key' parameter is an empty tuple
        - if any column present in 'expected_dataframe' is not present in 'available_dataframe'
        - if 'expected_dataframe' is empty
        - if 'primary_key' does not prove uniqueness for 'expected_dataframe'
    """
    # To Do list:
    #
    # validate parameter types                 :: Done
    # handle empty parameter values            :: Done
    # check all columns exist                  :: Done
    # strict check given primary key is valid  :: Done
    # spot & report missing primary_key        :: Done
    # spot & report duplicated primary_key     :: Done
    # spot & report mismatching attributes     :: Done
    # cache if necessary                       :: Done
    # repartition if necessary                 :: Done
    # have a way to clean up resource          :: Done

    # validate input parameter types
    _check_instance_type(expected_dataframe, DataFrame, 'expected_dataframe')
    _check_instance_type(available_dataframe, DataFrame, 'available_dataframe')
    _check_instance_type(primary_key, (tuple,list), 'primary_key')

    # handle empty parameter
    if not primary_key:
        raise Exception("Please provide valid non-empty tuple as Primary Key")

    # list down primary and non primary attributes
    primary_attributes = [ a.lower() for a in primary_key ]
    non_primary_attributes = [ col.lower()
                                   for col in expected_dataframe.columns
                                       if col.lower() not in primary_attributes
                             ]

    # check all columns exist
    # all columns on expected_dataframe, must be present in available_dataframe
    missing_attributes = set((*primary_attributes, *non_primary_attributes)).difference(
                                 { c.lower() for c in available_dataframe.columns }
                             )
    if len(missing_attributes) > 0:
        raise Exception("Few attributes are missing in second data frame : {}"\
                            .format(missing_attributes))

    # repartition both data frames on primary keys
    expected_dataframe = _repartition_df_with_keys(expected_dataframe, primary_attributes)
    available_dataframe = _repartition_df_with_keys(available_dataframe, primary_attributes)

    # cache the data frames, if it is not cached before
    uncache_list = []  # to clear only those data frames that were cached by this function
    for d_f in (expected_dataframe, available_dataframe):
        if not d_f.is_cached:
            d_f.persist()
            uncache_list.append(d_f)

    # handle empty parameter
    if not expected_dataframe.count():
        raise Exception("Input 'expected_dataframe' can not be empty")

    # strict check given primary key is valid
    if expected_dataframe.groupBy(*primary_attributes).count().where("count > 1").count():
        raise Exception("Given primary key is not unique for the first data frame. \
                         Please retry with valid primary key")

    # spot & report missing primary key
    df_missing_primary_key = _spot_missing_primary_key(
                                     expected_dataframe,
                                     available_dataframe,
                                     primary_attributes
                                 )

    # spot & report duplicated primary key
    df_duplicated_primary_key = _spot_duplicated_primary_key(
                                        expected_dataframe,
                                        available_dataframe,
                                        primary_attributes
                                    )

    temp_view_list = [] # to clear only the temp views that were created by this function

    # detect variance
    if len(non_primary_attributes) > 0 :
        # when few non primary columns exists, then explicitly identify their variances
        #
        # create temporary views
        expected_dataframe.createOrReplaceTempView("effulge_expected_view")
        temp_view_list.append("effulge_expected_view")
        available_dataframe.createOrReplaceTempView("effulge_available_view")
        temp_view_list.append("effulge_available_view")

        # spot & report mismatching attributes
        df_variance = _spot_mismatch_variance(
                              "effulge_expected_view",
                              "effulge_available_view",
                              primary_attributes,
                              non_primary_attributes
                          )
    else:
        # when non primary columns do not exists,
        # i.e, we only have primary columns,
        # then the variances are caught implicity with MISSING_PRIMARY_KEY check
        #
        df_variance = _get_empty_result_df(expected_dataframe, primary_attributes)
    #
    #
    # merge all variances
    effulge_variance_dataframe = df_variance.union(df_missing_primary_key)\
                                             .union(df_duplicated_primary_key)\
                                             .repartition(*primary_attributes)

    # persist output dataframe
    effulge_variance_dataframe.persist()
    effulge_variance_dataframe.count()

    # clear cache and temp views
    _clean_cache_and_view(uncache_list, temp_view_list)

    # Return the output dataframe
    return effulge_variance_dataframe


#
def summarize_variance(variance_df):
    """
    Function to summarize the variance dataframe.
    It will provide count and percentage information for each variance attribute group.

    Parameters:
        Name: variance_df
        Type:    pyspark.sql.dataframe.DataFrame

    Return Type:
        pyspark.sql.dataframe.DataFrame
        Format:

    +-----------------------------+--------------------------+-------------------------------+
    |  EFFULGE_VARIANCE_PROVOKER  |  EFFULGE_VARIANCE_COUNT  |  EFFULGE_VARIANCE_PERCENTAGE  |
    +-----------------------------+--------------------------+-------------------------------+
    |                             |                          |                               |
    |                             |                          |                               |
    +-----------------------------+--------------------------+-------------------------------+

    Throws Exception:
        - If the input dataframe does not contain 'EFFULGE_VARIANCE_PROVOKER' column
    """
    all_columns = [ col_name.upper() for col_name in variance_df.columns ]
    if "EFFULGE_VARIANCE_PROVOKER" not in all_columns:
        raise Exception("Input Dataframe does not have column 'EFFULGE_VARIANCE_PROVOKER'")
    total_variance_count = variance_df.count()
    # create summary dataframe
    summary_df = variance_df.groupBy(
                                "EFFULGE_VARIANCE_PROVOKER"
                            ).count()\
                            .withColumnRenamed(
                                "count",
                                "EFFULGE_VARIANCE_COUNT"
                            ).withColumn(
                                "EFFULGE_VARIANCE_PERCENTAGE",
                                pyspark_round(
                                    col("EFFULGE_VARIANCE_COUNT") * 100 / total_variance_count,
                                    2
                                )
                            ).orderBy(
                                "EFFULGE_VARIANCE_COUNT",
                                ascending=False
                            )
    # persist summary dataframe
    summary_df.persist()
    summary_df.count()
    # return summary dataframe
    return summary_df


#
def _spot_corrupted_attributes(row_object, primary_column_list,
                               suspect_column_list, left_prefix, right_prefix):
    """
    Function to compare column values and identify mismatching column names
        Reads the 'row_object'
            which has 1 attribute for each column name in the 'primary_column_list'
        and has 2 attributes for each column name in the 'suspect_column_list'
        - one of the attribute will be prefixed with 'left_prefix' string
        - and the other attribute will be prefixed with 'right_prefix' string
        Then it compares corresponding attribute pairs
        If values of any attribute pair are different,
            then it tracks that attribute name in a buffer list
        At the end, it return a Row containing primary attributes
        and an additional column with contents of buffer list.

    Parameters:
        Name: row_object
        Type:    pyspark.sql.types.Row
        Name: primary_column_list
        Type:    list of String
        Name: suspect_column_list
        Type:    list of String
        Name: left_prefix
        Type:    String
        Name: right_prefix
        Type:    String

    Return Type:
        pyspark.sql.types.Row
    """
    varying_attributes = { c for c in suspect_column_list
                                 if not _are_equal(
                                            row_object[left_prefix + c],
                                            row_object[right_prefix + c]
                                        )
                         }
    output_dict = { p:row_object[p] for p in primary_column_list }
    output_dict['EFFULGE_VARIANCE_PROVOKER'] = sorted(varying_attributes)
    return Row(**output_dict)


#
def _are_equal(var_1, var_2):
    """
    Function to compare two variables for equality.

    Parameters:
        Name: var_1
        Type:    Any
        Name: var_2
        Type:    Any

    Return Type:
        Boolean
    """
    if isinstance(var_1, (tuple, list, dict)):
        is_equal = compare_two_collections(var_1, var_2)
    else:
        is_equal = var_1 == var_2
    return is_equal


#
def _repartition_df_with_keys(dataframe_object, partition_columns_tuple):
    """
    Function to check and repartition a dataframe based on given partition columns.
        Firstly this function reads the Physical Query Plan for 'dataframe_object'
        and searches to find partition columns
        If partitioning columns exist and if it is same as 'partition_columns_typle'
            then the input 'dataframe_object' is returned without change
        But if partitioning doesn't exist
          OR if partitioning columns are different from 'partition_columns_tuple'
            then a new dataframe is created
            by repartitioning 'dataframe_object' on 'partition_columns_typle'
            and this new repartitioned dataframe is returned

    Parameters:
        Name: dataframe_object
        Type:    pyspark.sql.dataframe.DataFrame
        Name: partition_columns_tuple
        Type:    tuple of String

    Return Type:
        pyspark.sql.dataframe.DataFrame
    """
    current_dataframe_partitioner = []
    # pylint: disable=protected-access
    for line in dataframe_object._jdf.queryExecution().simpleString().split("\n"):
    # pylint: enable=protected-access
        if 'hashpartitioning' in line.lower():
            #
            current_dataframe_partitioner = [ x.split('#')[0]
                                                  for x in line.lower()\
                                                               .split('hashpartitioning(', 1)[1]\
                                                               .split(')', 1)[0]\
                                                               .split(' ')[0:-1]
                                            ]
            break
    #
    s_1 = set(partition_columns_tuple)
    s_2 = set(current_dataframe_partitioner)
    #
    if s_1.difference(s_2) or s_2.difference(s_1):
        dataframe_object = dataframe_object.repartition(*partition_columns_tuple)
    #
    return dataframe_object


#
def _check_instance_type(given_object, acceptable_classes, param_name):
    """
    Function to verify if 'given_object' is an instance of 'acceptable_classes'

    Parameters:
        Name: given_object
        Type:    any type
        Name: acceptable_classes
        Type:    a Class Name or a collection of Class Names
        Name: param_name
        Type:    String

    Return Type:
        None

    Throws Exception:
        - if "given_object" is not an instance of "acceptable_classes"
    """
    if not isinstance(given_object, acceptable_classes):
        raise Exception("Given input '{0}' :: {1} is invalid. Expected types are {2}"\
	                    .format(param_name, type(given_object), acceptable_classes))


#
def _spot_missing_primary_key(reference_df, received_df, prime_columns):
    """
    Function to identify the primary keys,
        which have a record in "reference_df"
        but do not have a record in "received_df"

    Parameters:
        Name: reference_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: received_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: prime_columns
        Type:    list/tuple of Strings

    Return Type:
        pyspark.sql.dataframe.DataFrame
    """
    try:
        df_missing_primary_attributes = reference_df.select(*prime_columns)\
                                            .subtract(
                                                received_df.select(*prime_columns)
                                            )
        _schema = df_missing_primary_attributes.schema
        _schema.add("EFFULGE_VARIANCE_PROVOKER", ArrayType(StringType()) )
        df_missing_primary_key = df_missing_primary_attributes\
                                     .rdd\
                                     .map(
                                         lambda r:
                                             Row(
                                                 **r.asDict(),
                                                 EFFULGE_VARIANCE_PROVOKER=["MISSING_PRIMARY_KEY"]
                                             )
                                     ).toDF( schema = _schema )
    except ValueError as exp:
        if str(exp) == "RDD is empty":
            # create empty result set
            df_missing_primary_key = _get_empty_result_df(reference_df, prime_columns)
        else:
            # raise the same exception, when ValueError message is different
            raise exp
    return df_missing_primary_key


#
def _spot_duplicated_primary_key(reference_df, received_df, prime_columns):
    """
    Function to identify the primary keys,
        which have only 1 record in "reference_df"
        but have more than 1 record in "received_df"

    Parameters:
        Name: reference_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: received_df
        Type:    pyspark.sql.dataframe.DataFrame
        Name: prime_columns
        Type:    list/tuple of Strings

    Return Type:
        pyspark.sql.dataframe.DataFrame
    """
    df_duplicated_primary_attributes = reference_df.select(*prime_columns)\
                                           .join(
                                               received_df.select(*prime_columns)
                                               , prime_columns
                                               , "inner"
                                           ).groupBy(*prime_columns).count()\
                                           .where("count > 1").select(*prime_columns)
    try:
        _schema = df_duplicated_primary_attributes.schema
        _schema.add("EFFULGE_VARIANCE_PROVOKER", ArrayType(StringType()) )
        df_duplicated_primary_key = df_duplicated_primary_attributes\
                                        .rdd\
                                        .map(
                                            lambda r:
                                                Row(
                                                    **r.asDict(),
                                                    EFFULGE_VARIANCE_PROVOKER=[
                                                        "DUPLICATE_PRIMARY_KEY"
                                                    ]
                                                )
                                        ).toDF( schema = _schema )
    except ValueError as exp:
        if str(exp) == "RDD is empty":
            # create empty result set
            df_duplicated_primary_key = _get_empty_result_df(reference_df, prime_columns)
        else:
            # raise the same exception, when ValueError message is different
            raise exp
    return df_duplicated_primary_key


#
def _spot_mismatch_variance(reference_view_name, received_view_name,
                                prime_columns, non_prime_columns):
    """
    Function to identify the mismatching records
    and also to identify the columns responsible for mismatch

    Parameters:
        Name: reference_view_name
        Type:    String
        Name: received_view_name
        Type:    String
        Name: prime_columns
        Type:    list/tuple of Strings
        Name: non_prime_columns
        Type:    list/tuple of Strings

    Return Type:
        pyspark.sql.dataframe.DataFrame
    """
    df_mismatch = SparkSession.getActiveSession().sql(
            """
            select {0} from {1}
            MINUS
            select {0} from {2}
            """.format(
                   ", ".join((*prime_columns, *non_prime_columns)),
                   reference_view_name,
                   received_view_name
               )
        ).select(prime_columns)

    # retains same column names for primary attributes
    # but renames non primary attributes to have "e_" prefix
    df_expected_with_renamed_columns = SparkSession.getActiveSession().sql(
            """
            select
                -- primary columns
                {},
                -- non primary columns with "e_" prefix
                {}
            from {}
            """.format(
                   ", ".join( prime_columns ),
                   ", ".join( [ "{0} as e_{0}".format(c) for c in non_prime_columns ] ),
                   reference_view_name
               )
        )

    # retains same column names for primary attributes
    # but renames non primary attributes to have "a_" prefix
    df_available_with_renamed_columns = SparkSession.getActiveSession().sql(
            """
            select
                -- primary columns
                {},
                -- non primary columns with "a_" prefix
                {}
            from {}
            """.format(
                   ", ".join( prime_columns ),
                   ", ".join( [ "{0} as a_{0}".format(c) for c in non_prime_columns ] ),
                   received_view_name
               )
        )

    df_mismatch_join = df_mismatch.join(
                                       df_expected_with_renamed_columns,
                                       prime_columns,
                                       "inner"
                                 ).join(
                                       df_available_with_renamed_columns,
                                       prime_columns,
                                       "inner"
                                 )

    # for each mismatch record, compare and identify variance columns
    try:
        df_variance = df_mismatch_join\
                          .rdd\
                          .map(
                              lambda r:
                                  _spot_corrupted_attributes(r, prime_columns, non_prime_columns,
                                                             "e_", "a_")
                          ).toDF()
    except ValueError as exp:
        if str(exp) == "RDD is empty":
            # create empty result set
            df_variance = _get_empty_result_df(df_expected_with_renamed_columns, prime_columns)
        else:
            # raise the same exception, when ValueError message is different
            raise exp
    return df_variance


#
def _clean_cache_and_view(cached_dataframe_list, temporary_view_list):
    """
    Function to explicitly free dataframe cache and to remove temporary views

    Parameters:
        Name : cached_dataframe_list
        Type :    list of pyspark.sql.dataframe.DataFrame objects
        Name : temporary_view_list
        Type :    list of String

    Return Type:
        None
    """
    # clear previously cached list
    for d_f in cached_dataframe_list:
        d_f.unpersist(blocking=True)

    # clear temporary view
    for view in temporary_view_list:
        SparkSession.getActiveSession().catalog.dropTempView(view)


#
def _get_empty_result_df(reference_df, primary_column_list):
    """
    Function to create an empty dataframe containing Primary Key columns and an additional EFFULGE_VARIANCE_PROVOKER column
    Parameters:
        Name : reference_df
        Type :     pyspark.sql.dataframe.DataFrame object
        Name : primary_column_list
        Type :     list of String
    Return Type:
        pyspark.sql.dataframe.DataFrame object
    """
    _schema = reference_df.select( *primary_column_list ).schema
    _schema.add("EFFULGE_VARIANCE_PROVOKER", ArrayType(StringType()) )
    #
    empty_df = SparkSession.getActiveSession().createDataFrame(
        SparkSession.getActiveSession().sparkContext.emptyRDD(),
        _schema
    )
    #
    return empty_df

